'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.createLeaf = exports.createTree = exports.createTreeObservable = exports.createTreeSetNextState = undefined;

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _rx = require('rx');

var _rx2 = _interopRequireDefault(_rx);

var _utils = require('../utils');

var _errors = require('../errors');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

var createTreeSetNextState = exports.createTreeSetNextState = function createTreeSetNextState(childrenObservable) {
  var newStateSubject = new _rx2.default.Subject();
  newStateSubject.withLatestFrom(childrenObservable).subscribe(function (_ref) {
    var _ref2 = _slicedToArray(_ref, 2);

    var newState = _ref2[0];
    var children = _ref2[1];

    var keys = Object.keys(children);
    var newKeys = Object.keys(newState);

    if (newKeys.length !== keys.length) {
      (0, _errors.throwShapeError)();
    }

    keys.forEach(function (key) {
      if (key in newState) {
        children[key].setNextState(newState[key]);
      } else {
        (0, _errors.throwShapeError)();
      }
    });
  });

  return function (newState) {
    newStateSubject.onNext(newState);
  };
};

var createTreeObservable = exports.createTreeObservable = function createTreeObservable(childrenObservable) {
  return childrenObservable.flatMapLatest(function (children) {
    var keys = Object.keys(children);
    if (keys.length) {
      var _Rx$Observable;

      return (_Rx$Observable = _rx2.default.Observable).combineLatest.apply(_Rx$Observable, _toConsumableArray(keys.filter(function (key) {
        return !children[key].provisional;
      }).map(function (key) {
        return children[key].combinedObservable;
      })).concat([function () {
        for (var _len = arguments.length, latestValues = Array(_len), _key = 0; _key < _len; _key++) {
          latestValues[_key] = arguments[_key];
        }

        return latestValues.reduce(function (acc, val, i) {
          acc[keys[i]] = val;
          return acc;
        }, {});
      }]));
    } else {
      return _rx2.default.Observable.just({});
    }
  });
};

var updateAddChildrenSubject = function updateAddChildrenSubject(initialState, addChildrenSubject, pauser) {
  var keys = Object.keys(initialState);
  pauser.onNext(false);
  keys.forEach(function (key) {
    addChildrenSubject.onNext({ key: key, value: initialState[key], provisional: false });
  });
  pauser.onNext(true);
};

var createFinalTreeFromProvisionalNode = function createFinalTreeFromProvisionalNode(_ref3) {
  var initialState = _ref3.initialState;
  var createNode = _ref3.createNode;
  var provisionalNode = _ref3.provisionalNode;
  var addChildrenSubject = provisionalNode.addChildrenSubject;
  var pauser = provisionalNode.pauser;

  updateAddChildrenSubject(initialState, addChildrenSubject, pauser);
  return createNode({
    provisional: false,
    provisionalNode: provisionalNode
  });
};

var createInitialTree = function createInitialTree(_ref4) {
  var initialState = _ref4.initialState;
  var pauser = _ref4.pauser;
  var hookMap = _ref4.hookMap;
  var createNode = _ref4.createNode;
  var provisional = _ref4.provisional;
  var provisionalNode = _ref4.provisionalNode;

  if (!pauser) {
    pauser = new _rx2.default.BehaviorSubject(true);
  }
  if (!hookMap) {
    hookMap = new WeakMap();
  }
  var addChildrenSubject = new _rx2.default.Subject();
  var getChildrenSubject = new _rx2.default.BehaviorSubject({});
  var childrenObservable = addChildrenSubject.withLatestFrom(getChildrenSubject, function (child, acc) {
    var key = child.key;
    var value = child.value;
    var provisional = child.provisional;

    var oldNode = acc[key] && acc[key].nodeSubject.getValue();
    if (!oldNode || oldNode.provisional) {
      var newNode = createTree({
        initialState: value,
        pauser: pauser,
        hookMap: hookMap,
        createNode: createNode,
        provisional: provisional,
        provisionalNode: oldNode
      });
      if (!oldNode) {
        return Object.assign({}, acc, _defineProperty({}, key, newNode));
      } else {
        return acc;
      }
    } else if (!oldNode.provisional) {
      (0, _errors.throwFinalizedError)(key);
    } else {
      return acc;
    }
  }).shareReplay(1);

  var valueSubject = new _rx2.default.ReplaySubject(1);

  childrenObservable.subscribe(getChildrenSubject);
  childrenObservable.subscribeOnError(function (err) {
    throw err;
  });

  if (!provisional) {
    updateAddChildrenSubject(initialState, addChildrenSubject, pauser);
  }

  var valueObservable = createTreeObservable(childrenObservable).pausable(pauser);
  var setNextState = createTreeSetNextState(childrenObservable, pauser);

  valueObservable.subscribe(valueSubject);

  var node = createNode({
    addChildrenSubject: addChildrenSubject,
    getChildrenSubject: getChildrenSubject,
    pauser: pauser,
    observable: valueSubject.asObservable(),
    hookMap: hookMap,
    setNextState: setNextState,
    provisional: provisional,
    provisionalNode: provisionalNode
  });

  return node;
};

var createTree = exports.createTree = function createTree(_ref5) {
  var initialState = _ref5.initialState;
  var pauser = _ref5.pauser;
  var hookMap = _ref5.hookMap;
  var createNode = _ref5.createNode;
  var provisional = _ref5.provisional;
  var provisionalNode = _ref5.provisionalNode;

  if (typeof initialState === 'undefined') {
    return createInitialTree({
      initialState: null,
      pauser: pauser,
      hookMap: hookMap,
      createNode: createNode,
      provisional: true
    });
  } else if (!provisional) {
    if (!(0, _utils.isPlainObject)(initialState)) {
      return createLeaf({
        initialState: initialState,
        pauser: pauser,
        createNode: createNode,
        hookMap: hookMap,
        provisionalNode: provisionalNode
      });
    } else if (provisionalNode) {
      return createFinalTreeFromProvisionalNode({
        initialState: initialState,
        createNode: createNode,
        provisionalNode: provisionalNode
      });
    }
  }

  return createInitialTree({
    initialState: initialState,
    pauser: pauser,
    hookMap: hookMap,
    createNode: createNode,
    provisional: provisional,
    provisionalNode: provisionalNode
  });
};

var createLeaf = exports.createLeaf = function createLeaf(_ref6) {
  var initialState = _ref6.initialState;
  var pauser = _ref6.pauser;
  var createNode = _ref6.createNode;
  var hookMap = _ref6.hookMap;
  var provisionalNode = _ref6.provisionalNode;

  if (provisionalNode && Object.keys(provisionalNode.getChildrenSubject.getValue()).length) {
    throw new Error('Tried to create leaf node when provisional has children');
  }
  if (!pauser) {
    pauser = new _rx2.default.BehaviorSubject(true);
  }
  if (!hookMap) {
    hookMap = new WeakMap();
  }
  var subject = new _rx2.default.BehaviorSubject(initialState);
  var observable = subject.distinctUntilChanged();
  var setNextState = function setNextState(newState) {
    subject.onNext(newState);
  };

  return createNode({
    observable: observable,
    hookMap: hookMap,
    pauser: pauser,
    setNextState: setNextState,
    provisional: false,
    provisionalNode: provisionalNode
  });
};

exports.default = createTree;